# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['titan', 'titan.exposures', 'titan.features', 'titan.interactions']

package_data = \
{'': ['*'],
 'titan': ['params/*',
           'settings/atlanta/*',
           'settings/chicago/*',
           'settings/mississippi/*',
           'settings/nyc-msm/*',
           'settings/scott/*']}

install_requires = \
['nanoid>=2.0,<3.0',
 'networkx>=2.4,<3.0',
 'numpy>=1.18,<2.0',
 'paraml>=0.1,<0.2']

extras_require = \
{'docs': ['mkdocs>=1.1,<2.0',
          'mkdocs-material>=7.0,<8.0',
          'mkdocstrings>=0.15.0,<0.16.0'],
 'linting': ['black>=20.8b1,<21.0', 'flake8>=3.8,<4.0', 'mypy>=0.812,<0.813']}

entry_points = \
{'console_scripts': ['run_titan = titan:script_init']}

setup_kwargs = {
    'name': 'titan-model',
    'version': '2.5.0',
    'description': 'TITAN Agent Based Model',
    'long_description': '# TITAN Simulation\n[![DOI](https://zenodo.org/badge/80315242.svg)](https://zenodo.org/badge/latestdoi/80315242)\n[![GitHub release (latest by date)](https://img.shields.io/github/v/release/pph-collective/TITAN)](https://github.com/pph-collective/TITAN/releases/latest/) [![](https://github.com/pph-collective/TITAN/workflows/Unit%20Tests/badge.svg)](https://github.com/pph-collective/TITAN/actions) [![codecov](https://codecov.io/gh/pph-collective/TITAN/branch/develop/graph/badge.svg?token=wjkExshhyh)](https://codecov.io/gh/pph-collective/TITAN) [![GitHub](https://img.shields.io/github/license/pph-collective/TITAN)](https://github.com/pph-collective/TITAN/blob/develop/LICENSE) [![Stable](https://img.shields.io/badge/docs-stable-blue.svg)](https://pph-collective.github.io/TITAN/)\n\nTITAN (Treatment of Infectious Transmissions through Agent-based Network) is an agent-based simulation model used to explore contact transmission in complex social networks. Starting with the initializing agent population, TITAN iterates over a series of stochastic interactions where agents can interact with one another, transmit infections through various medium, and enter and exit the care continuum. The purpose of TITAN is to evaluate the impact of prevention and treatment models on incidence and prevalence rates of the targeted disease(s) through the use of data fitting simulated trajectories and rich statistics of primary/sub-population attributable proportions.\n\nAgent populations are defined as graphs (nodes connected by edges). Nodes in the graph are used to represent the attributes (or collection of attributes) of an agent (person), and edges define the type of relationship between agents. In practice, a graph represents a social network of connected people through various relationship types, and provides the medium for which agents can interact.\n\n## Getting Started\n\nInstall the package\n\n```\npip install titan-model\n```\n\nThis includes the script `run_titan` which can then be used to run the model.\n\n### Prerequisites\n\n* Python (or pypy) 3.6 or later\n\n## Running the Model\n\n```\nrun_titan -p my_params.yml\n```\n\nTo run the model, execute the `run_titan` program. See [TITAN params](https://pph-collective.github.io/titan-params-app) for documentation on how to set and use parameters.\n\nResults of the model are generated and aggregated into the `/results/` directory by default. If the model is re-run, the existing results will be overwritten.\n\n\n## Built With\n* [Python3.x](https://www.python.org/downloads/release/python-374/) - Programming language\n\n  Van Rossum G, Drake FL. Python 3 Reference Manual. Scotts Valley, CA: CreateSpace; 2009.\n\n* [Networkx](https://networkx.github.io/) - Network structure backend\n\n  Hagberg A, Swart P, S Chult D. Exploring network structure, dynamics, and function using NetworkX. 2008.\n\n* [Numpy](http://www.numpy.org/) - Numerical libraries\n\n  Oliphant TE. A guide to NumPy. Vol. 1. Trelgol Publishing USA; 2006.\n\n## Authors\n\n* **Lars Seeman** - *Initial work*\n* **Max King** - *Continued development*\n* **Sarah Bessey** - *Continued development*\n* **Mary McGrath** - *Continued development*\n\n## License\n\nThis project is licensed under the GNU General Public License version 3 - see the [LICENSE.md](LICENSE.md) file for details\n',
    'author': 'Sam Bessey',
    'author_email': 'sam_bessey@brown.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://titanmodel.org',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from . import utils
from .location import Location

# ================ CORE PROBABILITIES ========================


@utils.memo
def get_death_rate(
    hiv: bool,
    aids: bool,
    drug_type: str,
    sex_type: str,
    haart_adh: int,
    race: str,
    location: Location,
    steps_per_year: int,
) -> float:
    """
    Find the death rate of an agent given a set of attributes.

    args:
        hiv: whether the agent is HIV+
        aids: whether the agent has AIDS
        drug_type: whether the PWID base death rate should be used or the base one
        haart_adh: whether an agent is haart adherent
        race: the race of the agent
        location: agent's location
        steps_per_year: the number of model steps in a year

    returns:
        the probability of an agent with these characteristics dying in a given time step
    """
    param = location.params.demographics

    death_param = param[race].sex_type[sex_type].drug_type[drug_type].death_rate

    p = death_param.base

    if aids:
        p *= death_param.aids
    elif hiv:
        if haart_adh:
            p *= death_param.haart_adherent
        else:
            p *= death_param.hiv

    # putting it into per 1 person-month from per 1000 person years
    return p / (steps_per_year * 1000.0)



import os
import csv
from typing import Dict, Any
from shutil import make_archive, unpack_archive
from tempfile import mkdtemp
import glob
import re
import logging

from .population import Population
from .agent import Agent, Relationship
from .parse_params import ObjMap
from .location import Location
from . import features
from . import exposures
from . import utils

agent_feature_attrs = [
    feature.name for feature in features.BaseFeature.__subclasses__()
]
agent_exposure_attrs = [
    exposure.name for exposure in exposures.BaseExposure.__subclasses__()
]

# these are functionally saved in the relationships or other files and complicate the agent file
agent_exclude_attrs = (
    {"partners", "relationships"}.union(agent_feature_attrs).union(agent_exposure_attrs)
)


def write(pop: Population, dir: str, compress: bool = True) -> str:
    """
    Write a non-empty Population to file.

    args:
        pop: a non-empty agent population
        dir: path to directory where files should be written
        compress: whether to compress and archive the csv

    returns:
        path, or archive name if compress is true
    """
    assert len(pop.relationships) > 0, "Can't write empty population"

    utils.set_up_logging(pop.params)

    # open agent file
    agent_file = os.path.join(dir, f"{pop.id}_agents.csv")

    a = next(iter(pop.all_agents))
    # get all attributes
    agent_attrs = [k for k in a.__dict__.keys() if k not in agent_exclude_attrs]

    write_class_file(agent_file, pop.all_agents, agent_attrs)

    extra_files = []

    # write agent extras (features, exposures) to their own files
    def write_extra_class(extra_attrs, extra_type):
        for extra in extra_attrs:
            extra_obj = getattr(a, extra)
            extra_attrs = list(extra_obj.__dict__.keys())
            extra_file = os.path.join(dir, f"{pop.id}_{extra_type}_{extra}.csv")
            extra_files.append(extra_file)
            write_extra_class_file(extra_file, pop.all_agents, extra, extra_attrs)

    write_extra_class(agent_feature_attrs, "feat")
    write_extra_class(agent_exposure_attrs, "exposure")

    # open relationship file
    rel_file = os.path.join(dir, f"{pop.id}_relationships.csv")

    r = next(iter(pop.relationships))
    rel_attrs = list(r.__dict__.keys())

    write_class_file(rel_file, pop.relationships, rel_attrs)

    if compress:
        archive_name = make_archive(
            os.path.join(dir, f"{pop.id}_pop"), "gztar", root_dir=dir, base_dir="."
        )
        os.remove(agent_file)
        os.remove(rel_file)
        for f in extra_files:
            os.remove(f)

        return archive_name
    else:
        return dir


def write_extra_class_file(file_name, collection, extra, attrs):
    logging.info(f"Creating {file_name}")
    with open(file_name, "w", newline="") as f:
        writer = csv.DictWriter(f, fieldnames=attrs)
        writer.writeheader()
        for item in collection:
            feat = getattr(item, extra)
            writer.writerow({attr: repr(getattr(feat, attr)) for attr in attrs})


def write_class_file(file_name, collection, attrs):
    logging.info(f"Creating {file_name}")
    with open(file_name, "w", newline="") as f:
        writer = csv.DictWriter(f, fieldnames=attrs)
        writer.writeheader()
        for item in collection:
            writer.writerow({attr: repr(getattr(item, attr)) for attr in attrs})


def read(params: ObjMap, path: str) -> Population:
    """
    Read a population from file and return a Population instance

    args:
        params: the parameters used for creating this popultation
        path: path where [id]_agents.csv and [id]_relationships.csv are or tar.gz file
            containing population
    returns:
        the re-constituted population
    """
    if os.path.isfile(path):
        dir = mkdtemp()
        unpack_archive(path, dir)
        path = dir

    agent_file = glob.glob(os.path.join(path, "*_agents.csv"))[0]
    rel_file = glob.glob(os.path.join(path, "*_relationships.csv"))[0]
    feat_files = glob.glob(os.path.join(path, "*_feat_*.csv"))
    exposure_files = glob.glob(os.path.join(path, "*_exposure_*.csv"))
    assert os.path.isfile(agent_file), f"can't find agents.csv in {dir}"
    assert os.path.isfile(rel_file), f"can't find relationships.csv in {dir}"

    _, agent_filename = os.path.split(agent_file)
    id = agent_filename[:8]

    # create feature dict
    agent_extras: Dict[str, Dict] = {}

    def update_agent_extras(files, extra_type):
        pattern = re.compile(f"^.*_{extra_type}_(.*)\\.csv$")
        for file in files:
            m = pattern.match(file)
            if m is not None:
                extra = m.group(1)
                agent_extras[extra] = {}
                with open(file, newline="") as f:
                    reader = csv.DictReader(f)
                    for row in reader:
                        agent_extras[extra][int(row["agent"])] = row

    update_agent_extras(feat_files, "feat")
    update_agent_extras(exposure_files, "exposure")

    # don't create any agents on init
    params.model.num_pop = 0
    pop = Population(params, id=id)

    # re-create all agents and add to population
    with open(agent_file, newline="") as f:
        reader = csv.DictReader(f)
        for row in reader:
            a = create_agent(
                row,
                params.classes.bond_types.keys(),
                pop.geography.locations,
                agent_extras,
            )
            pop.add_agent(a)

    # update num_pop to actual population
    params.model.num_pop = pop.all_agents.num_members()

    # re-create all relationships and add to population
    with open(rel_file, newline="") as f:
        reader = csv.DictReader(f)
        for row in reader:
            r = create_relationship(row, pop)
            pop.add_relationship(r)

    pop.update_agent_components()

    return pop


def create_agent(
    row: Dict[str, str],
    bond_types,
    locations: Dict[str, Location],
    agent_extras: Dict[str, Any],
) -> Agent:
    """
    Initialize an Agent from a row of the saved population
    """
    init_attrs = ["sex_type", "age", "race", "drug_type", "id", "location"]
    location = locations[eval(row["location"])]
    agent = Agent(
        eval(row["sex_type"]),
        eval(row["age"]),
        eval(row["race"]),
        eval(row["drug_type"]),
        location,
        eval(row["id"]),
    )

    for attr, val in row.items():
        if attr not in init_attrs:
            setattr(agent, attr, eval(val))

    for extra in agent_extras:
        extra_row = agent_extras[extra][agent.id]
        agent_extra = getattr(agent, extra)
        for attr, val in extra_row.items():
            if not attr == "agent":
                setattr(agent_extra, attr, eval(val))

        if agent_extra.active:
            agent_extra.add_agent(agent)

    agent.partners = {bond: set() for bond in bond_types}

    return agent


def create_relationship(row: Dict[str, str], pop: Population) -> Relationship:
    """
    Initialize a Relationship from a row of the saved population
    """
    init_attrs = ["agent1", "agent2", "duration", "bond_type", "id"]
    agent1 = find_agent(pop, row["agent1"])
    agent2 = find_agent(pop, row["agent2"])
    rel = Relationship(
        agent1,
        agent2,
        eval(row["duration"]),
        eval(row["bond_type"]),
        id=eval(row["id"]),
    )

    for attr, val in row.items():
        if attr not in init_attrs:
            setattr(rel, attr, eval(val))

    return rel


def find_agent(pop: Population, id_str: str) -> Agent:
    """
    Given a Population and an id (as a string), return the Agent with that id
    """
    id = eval(id_str)
    for a in pop.all_agents:
        if a.id == id:
            return a

    raise Exception(f"Agent {id_str} not found")



from typing import List, Dict

from .. import agent
from .. import population
from .. import model


class BaseFeature:
    """
    Interface class for an agent-oriented feature of the TITAN model.  It is intended to be
    inherited by feature classes to ensure that the expected methods/fields exist.

    The class takes advantage of both instance and class methods and attributes. The instance
    methods/attributes are used at the agent level, whereas the class methods/attributes are
    used at the population level.

    A class method can be called on an instance of an object, but it doesn't have access
    to that instance (e.g. agent.feature.add_agent has to be passed agent because it is a class
    method).
    """

    name: str = ""
    """Name of feature in the params file.  Also used to name the attribute in Agent"""

    stats: List[str] = []
    """List of names of stats that come from this feature (e.g. numFeat)"""

    def __init__(self, agent: "agent.Agent"):
        """
        Constructor for an instance of the feature.  This is called from within `Agent.__init__` and passes the agent to the feature to create a two way binding.  All features must have the attributes of `active` and `agent`.  By default `active` is false and `agent` is the passed agent.

        Called on all features whether or not param is enabled to make sure that references to other features within a feature do not cause errors (e.g. incar referring to prep, but prep isn't on).

        args:
            agent: The agent this feature instance is attached to.
        """
        self.active = False
        self.agent = agent

    @classmethod
    def init_class(cls, params):
        """
        Initialize any class level attributes (such as setting counters to zero). Called on every active feature on population initialization.

        args:
            params: parameters for this population
        """
        pass

    def init_agent(self, pop: "population.Population", time: int):
        """
        Initialize the agent for this feature during population initialization (`Population.create_agent`).  Called on only features that are enabled per the params.

        args:
            pop: the population this agent is a part of
            time: the current time step
        """
        pass

    def update_agent(self, model: "model.TITAN"):
        """
        Update the agent for this feature for a time step.  Called once per time step in `TITAN.update_all_agents`. Agent level updates are done after population level updates.   Called on only features that are enabled per the params.

        args:
            model: the instance of TITAN currently being run
        """
        pass

    @classmethod
    def add_agent(cls, agent: "agent.Agent"):
        """
        Add an agent to the class (not instance).  This can be useful if tracking population level statistics or groups, such as counts or newly active agents.

        This method is not called from anywhere in the model, but creates a cohesive api with `remove_agent`, which is called from `Population.remove_agent`.

        args:
            agent: the agent to add to the class attributes
        """
        pass

    @classmethod
    def remove_agent(cls, agent: "agent.Agent"):
        """
        Remove an agent from the class (not instance).  This can be useful if tracking population level statistics or groups, such as counts.

        This method is called from `Population.remove_agent`, but may also need to be called within the feature if an agent transitions from `active == True` to `active == False`.

        args:
            agent: the agent to remove from the class attributes
        """
        pass

    @classmethod
    def update_pop(cls, model: "model.TITAN"):
        """
        Update the feature for the entire population (class method).  This is useful for initializing class level trackers that need to be reset each time step, or if enabling a feature for agents needs to be evaluated within the context of the full population (limited slots, or similar).

        This is called in `TITAN.update_all_agents` before agent-level updates are made.

        args:
            model: the instance of TITAN currently being run
        """
        pass

    def set_stats(self, stats: Dict[str, int], time: int):
        """
        Update the `stats` dictionary passed for this agent.  Called from `output.get_stats` for each enabled feature in the model.

        The stats to be updated must be declared in the class attribute `stats` to make sure the dictionary has the expected keys/counter value initialized.

        args:
            stats: the dictionary to update with this agent's feature statistics
            time: the time step of the model when the stats are set
        """
        pass

    def get_acquisition_risk_multiplier(self, time: int, interaction_type: str):
        """
        Get a multiplier for how this feature affects acquisition of HIV for the given interaction_type.

        By default, returns 1.0

        args:
            time: the current model time step
            interaction_type: The type of interaction where the agent could acquire HIV (e.g. 'sex', 'injection' - from [params.classes.interaction_types])
        """
        return 1.0

    def get_transmission_risk_multiplier(self, time: int, interaction_type: str):
        """
        Get a multiplier for how this feature affects transmission of HIV for the given interaction_type.

        By default, returns 1.0

        args:
            time: the current model time step
            interaction_type: The type of interaction where the agent could transmit HIV (e.g. 'sex', 'injection' - from [params.classes.interaction_types])
        """
        return 1.0



from . import base_feature
from .. import agent
from .. import population
from .. import model


class ExternalExposure(base_feature.BaseFeature):

    name = "external_exposure"

    def __init__(self, agent: "agent.Agent"):
        super().__init__(agent)

        self.active = False

    def init_agent(self, pop: "population.Population", time: int):
        """
        Initialize the agent for this feature during population initialization (`Population.create_agent`).  Called on only features that are enabled per the params.

        If an agent has defined sex_type, with a random probability from params, assign them to be an agent with external exposure.

        args:
            pop: the population this agent is a part of
            time: the current time step
        """
        params = self.agent.location.params.external_exposure
        if self.agent.sex_type == params.sex_type:
            if pop.pop_random.random() < params.init:
                self.active = True

    def update_agent(self, model: "model.TITAN"):
        """
        Update the agent for this feature for a time step.  Called once per time step in `TITAN.update_all_agents`. Agent level updates are done after population level updates.   Called on only features that are enabled per the params.

        If the agent has external exposure, with a probability from params, convert the agent.

        args:
            model: the instance of TITAN currently being run
        """
        params = self.agent.location.params.external_exposure
        if self.active and model.run_random.random() < params.convert_prob:
            agent_exposure = getattr(self.agent, params.exposure)
            agent_exposure.convert(model)



import logging

from . import base_feature
from .. import model as hiv_model
from .. import utils


class SyringeServices(base_feature.BaseFeature):

    name = "syringe_services"

    enrolled_risk = 0.0

    def __init__(self, agent):
        super().__init__(agent)

        self.active = False

    @classmethod
    def init_class(cls, params):
        """
        Initialize enrolled risk to 0.

        args:
            params: the population params
        """
        cls.enrolled_risk = 0.0

    @classmethod
    def update_pop(cls, model: "hiv_model.TITAN"):
        """
        Update the feature for the entire population (class method).  This is useful for initializing class level trackers that need to be reset each time step, or if enabling a feature for agents needs to be evaluated within the context of the full population (limited slots, or similar).

        Enroll PWID agents in syringe services according to the syring_services timeline and params.

        args:
            model: the instance of TITAN currently being run
        """
        logging.info(("\n\n!!!!Engaging syringe services program"))
        ssp_num_slots = 0
        ssp_agents = {
            agent for agent in model.pop.pwid_agents if agent.syringe_services.active  # type: ignore[attr-defined]
        }

        for item in model.params.syringe_services.timeline.values():
            if item.start_time <= model.time < item.stop_time:
                cls.enrolled_risk = item.risk

                # linearly interpolate slots between start and stop
                ssp_num_slots = (item.num_slots_stop - item.num_slots_start) / (
                    item.stop_time - item.start_time
                ) * (model.time - item.start_time) + item.num_slots_start

                # If cap indicates all or no agents, do not change
                # otherwise, find true number of slots through distribution
                num_pwid_agents = model.pop.pwid_agents.num_members()
                if 0 < ssp_num_slots < num_pwid_agents:
                    ssp_num_slots = round(
                        model.run_random.betavariate(
                            ssp_num_slots,
                            num_pwid_agents - ssp_num_slots,
                        )
                        * num_pwid_agents
                    )
                break

        target_set = utils.safe_shuffle(
            (model.pop.pwid_agents.members - ssp_agents), model.run_random
        )

        # unenroll agents if above cap
        for agent in ssp_agents.copy():
            if len(ssp_agents) > ssp_num_slots:
                agent.syringe_services.active = False  # type: ignore[attr-defined]
                ssp_agents.remove(agent)
            else:
                break

        # enroll agents if below cap
        for agent in target_set:
            if len(ssp_agents) < ssp_num_slots:
                agent.syringe_services.active = True  # type: ignore[attr-defined]
                ssp_agents.add(agent)
            else:
                break

        logging.info(
            f"SSP has {ssp_num_slots} target slots with "
            f"{len(ssp_agents)} slots filled"
        )



from numpy import log  # type: ignore

"""
This file contains distributions that don't exist in numpy.
"""


def set_value(np_random, value):
    """
    A distribution that always returns the value passed

    args:
        np_random: random number generator (to conform to distribution interface)
        value: value to return
    """
    return value


def pert(np_random, low, peak, high, temperature):
    """
    A pert distribution, inspired by [tensorflow](https://github.com/tensorflow/probability/blob/c833ee5cd9f60f3257366b25447b9e50210b0590/tensorflow_probability/python/distributions/pert.py#L137)

    arguments must be so that:

    * low < peak < high
    * temperature > 0

    The support is `[low, high]`.  The `peak` must fit in that interval:
      `low < peak < high`.  The `temperature` is a positive parameter that
      controls the shape of the distribution. Higher values yield a sharper peak.

    args:
        np_random: random number generator (used to get beta)
        low: distribution low value
        peak: modal point in distribution
        high: distribution high value
        temperature: scaling factor
    """
    assert low < peak < high
    assert temperature > 0

    scale = high - low
    alpha = 1.0 + temperature * (peak - low) / scale
    beta = 1.0 + temperature * (high - peak) / scale
    return low + scale * np_random.beta(alpha, beta)


def weibull_modified(np_random, shape, scale):
    """
    Modified version of numpy's (single parameter) weibull distribution to use the 2-parameter weibull.

    args:
        np_random: random number generator
        shape: weibull shape parameter
        scale: weibull scale parameter
    """
    random_number = np_random.random()
    return scale * (-log(1 - random_number)) ** (1 / shape)
